#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include "audit_record.h"
#include "gnfalib.h"


typedef struct token_tab {
	char tok_id;
	u_short base_id;
	char *tok_addr;
} tok_ent_t;

#define 	TAB_LEN 	35

tok_ent_t token_table[TAB_LEN] = {
/* 0  */	 {AUT_HEADER		, 1  	, 0}	/* 0x14 */
/* 1  */	,{AUT_SOCKET 		, 100 	, 0}	/* 0x2E */
/* 2  */	,{AUT_ARG		, 200 	, 0}	/* 0x2D */
/* 3  */	,{AUT_ARG		, 205 	, 0}	/* 0x2D */
/* 4  */	,{AUT_ARG		, 210 	, 0}	/* 0x2D */
/* 5  */	,{AUT_ARG		, 215 	, 0}	/* 0x2D */
/* 6  */	,{AUT_EXEC_ENV		, 300 	, 0}	/* 0x3D */
/* 7  */	,{AUT_ATTR		, 400 	, 0}	/* 0x31 */
/* 8  */	,{AUT_DATA		, 500 	, 0}	/* 0x21 */
/* 9  */	,{AUT_EXIT		, 600 	, 0}	/* 0x11 */
/* 10  */	,{AUT_OTHER_FILE	, 700	, 0}	/* 0x21 */
/* 11  */	,{AUT_GROUPS		, 800	, 0}	/* 0x34 */
/* 12  */	,{AUT_IPC_PERM		, 900	, 0}	/* 0x32 */
/* 13  */	,{AUT_IPC		, 1000	, 0}	/* 0x22 */
/* 14  */	,{AUT_IP		, 1100	, 0}	/* 0x2B */
/* 15  */	,{AUT_IPORT		, 1200	, 0}	/* 0x2C */
/* 16  */	,{AUT_INVALID		, 1300	, 0}	/* 0x00 */
/* 17  */	,{AUT_OPAQUE		, 1400	, 0}	/* 0x29 */
/* 18  */	,{AUT_PATH		, 1500	, 0}	/* 0x23 */
/* 19  */	,{AUT_PATH		, 1503	, 0}	/* 0x23 */
/* 20  */	,{AUT_PATH		, 1506	, 0}	/* 0x23 */
/* 21  */	,{AUT_PATH		, 1509	, 0}	/* 0x23 */
/* 22  */	,{AUT_PROCESS		, 1600	, 0}	/* 0x26 */
/* 23  */	,{AUT_SERVER		, 1700	, 0}	/* 0x25 */
/* 24  */	,{AUT_SEQ		, 1800	, 0}	/* 0x2F */
/* 25  */	,{AUT_IN_ADDR		, 1900	, 0}	/* 0x2A */
/* 26  */	,{AUT_EXEC_ARGS		, 2000	, 0}	/* 0x3C */
/* 27  */	,{AUT_SUBJECT		, 2100	, 0}	/* 0x24 */
/* 28  */	,{AUT_TEXT		, 2200	, 0}	/* 0x28 */
/* 29  */	,{AUT_TEXT		, 2203	, 0}	/* 0x28 */
/* 30  */	,{AUT_TEXT		, 2206	, 0}	/* 0x28 */
/* 31  */	,{AUT_TEXT		, 2209	, 0}	/* 0x28 */
/* 32  */	,{AUT_RETURN		, 2300	, 0}	/* 0x27 */
/* 33  */	,{AUT_TRAILER		, 2400	, 0}	/* 0x13 */
/* 34  */	,{AUT_NEWGROUPS		, 2500	, 0}	/* 0x3B */
};

lookup_tok(char id, tok_ent_t *tab, int len)
{
	int i = 0;
	int found = 0;

	while (!found && i < len) 
		if (tab[i++].tok_id == id) 
			found = 1;
	if (!found) {
		fprintf(stderr, "unknown token id %d", AUT_OTHER_FILE);
		exit(1);
	}
	return i-1;
} 

nadfconv(register void *prec, int *pbuflen, char **pbuf)
{
    int N0, N = 1;                     /* audit data id */
    char *alpha = *pbuf + sizeof(int);/* current position in NADF buffer */
    u_short *IDL;
    struct au_token rec;                            /* native record */
    int i1;
    int ttllen;
    int len;
    short shtlen;
    u_short ushtlen;
    int size;
    int tok_id = *(char *) prec; 
    char *s;
    int k; 
    register char *crt_pos = (char *) prec;
    int arg_nb=0, text_nb=0, path_nb=0;	/* # repetitive tokens */
    int TOTAL;

    /* zeroing the table */

    for (k=0; k<TAB_LEN; k++)
	token_table[k].tok_addr = 0;

	if (tok_id == AUT_OTHER_FILE) {
	
	/* file token which begins every log file */
 
 	memcpy(&shtlen, (char *) prec+1+4+4, 2);
	ttllen = shtlen + 1+4+4+2;
    } else if (tok_id != AUT_HEADER) {
	fprintf(stderr, "expecting a header token while finding %d", tok_id);
	exit(1);
    } else {
    	memcpy(&ttllen, (char *) prec + 1, sizeof(long));
    }
    TOTAL = ttllen;
#ifdef DEBUG
    fprintf(stdout, "parsing record:");
#endif
    do {
    tok_id = *crt_pos;
#ifdef DEBUG
    fprintf(stdout, " 0x%x", tok_id);
#endif
    switch(tok_id) {

        case (0x2D):
	    if (arg_nb <= 3) {	

		/* accept max. 4 such tokens and skip over the rest  */

	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 2;
	    	token_table[k+arg_nb].tok_addr = crt_pos;
	    }  	
	    memcpy(&ushtlen, crt_pos+1+1+4, 2);
	    ttllen  -= 1 + 1+4+2 + ushtlen;
	    crt_pos += 1 + 1+4+2 + ushtlen;
	    arg_nb++;
            break;

        case (0x31):

	    k = lookup_tok(tok_id, token_table, TAB_LEN);
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4+4+4+4+4+4;
	    crt_pos += 1 + 4+4+4+4+4+4;
            break;
	    
        case (0x21):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 10;
	    token_table[k].tok_addr = crt_pos; 	
            switch(*(crt_pos+1+1)) {
                case (AUR_BYTE):
   	    	    ttllen  -= 1 + 1+1+1 + 1*(*(crt_pos+1+1+1));
		    crt_pos += 1 + 1+1+1 + 1*(*(crt_pos+1+1+1));
                    break;
               case (AUR_SHORT):
		    N = 18;
		    memcpy(&ushtlen, crt_pos +1+1+1, 2);
 	    	    ttllen  -= 1+1+1+1 + 2*ushtlen;
		    crt_pos += 1+1+1+1 + 2*ushtlen;
                    break;
                case (AUR_INT):
                case (AUR_LONG):
		    memcpy(&ushtlen, crt_pos +1+1+1, 2);
  	    	    ttllen  -= 1+1+1+1 + 4*ushtlen;
		    crt_pos += 1+1+1+1 + 4*ushtlen;
                   break;
            }
            break;

        case (0x52):
	    k = lookup_tok(tok_id, token_table, TAB_LEN);
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4+4;
	    crt_pos += 1 + 4+4; 
            break;

        case (0x11):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 9;
	    token_table[k].tok_addr = crt_pos; 	
	    memcpy(&ushtlen, crt_pos+1+4+4, 2);
	    ttllen  -= 1 + 4+4+2 + ushtlen;
	    crt_pos += 1 + 4+4+2 + ushtlen; 
            break;

        case (0x34):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 11;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 16*4;
	    crt_pos += 1 + 16*4; 
            break;

       case (0x3B):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 34;
	    token_table[k].tok_addr = crt_pos; 	
	    memcpy(&ushtlen, crt_pos + 1, 2);
	    ttllen  -= 1 + 2 + 4*ushtlen;
	    crt_pos += 1 + 2 + 4*ushtlen; 
            break;

        case (0x14):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 0;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4+1+2+2+4+4;
	    crt_pos += 1 + 4+1+2+2+4+4; 
            break;

        case (0x32):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 12;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4+4+4+4+4+4+4 + 4*4;
	    crt_pos += 1 + 4+4+4+4+4+4+4 + 4*4; 
            break;

        case (0x22): /* AUT_IPC */
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 13;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 +1+ 4;
	    crt_pos += 1 +1+ 4; 
            break;

        case (0x2C): /* == AUT_IPORT (Internet port address) */
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 15;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4;
	    crt_pos += 1 + 4; 
            break;

        case (0x00):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 16;
	    token_table[k].tok_addr = crt_pos; 	
	    memcpy(&ushtlen, crt_pos + 1, 2);
/*	    if (ushtlen > 250) {
		char *s;
		FILE *fd;

		fd = fopen("record", "w");
		fwrite(prec, 1, TOTAL, fd);
		fclose(fd);
		for (s=(char *) prec; s - (char *)prec != TOTAL; s++) {
		   fprintf(stderr, "%x ", *s);
	           
		}
	    	fprintf(stderr, "\n");
		exit(0);
	    } 
*/
	    ttllen  -= 1 + 2 + ushtlen+1;
	    crt_pos += 1 + 2 + ushtlen+1; 
            break;

        case (0x29):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 17;
	    token_table[k].tok_addr = crt_pos; 	
	    memcpy(&ushtlen, crt_pos + 1, 2);
	    ttllen  -= 1 + 2 + ushtlen;
	    crt_pos += 1 + 2 + ushtlen; 
            break;

        case (0x23):
	    if (path_nb <= 3) {	

		/* accept max. 4 such tokens and skip over the rest  */

	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 18;
	    	token_table[k+path_nb].tok_addr = crt_pos; 
	    }  	
	    memcpy(&ushtlen, crt_pos + 1, 2);
	    ttllen  -= 1 + 2 + ushtlen;
	    crt_pos += 1 + 2 + ushtlen; 
	    path_nb++;
            break;

        case (0x26):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 22;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4+4+4+4+4+4+4+4+4;
	    crt_pos += 1 + 4+4+4+4+4+4+4+4+4; 
            break;

        case (0x27):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 32;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 1+4;
	    crt_pos += 1 + 1+4; 
            break;

        case (0x25):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 23;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4+4+4+4+4;
	    crt_pos += 1 + 4+4+4+4+4; 
            break;

        case (0x2F):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 24;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4;
	    crt_pos += 1 + 4; 
            break;

         case (0x24):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 27;
	    token_table[k].tok_addr = crt_pos; 	
	    ttllen  -= 1 + 4+4+4+4+4+4+4+4+4;
	    crt_pos += 1 + 4+4+4+4+4+4+4+4+4; 
            break;

        case (0x3C): /* AUT_EXEC_ARGS */
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 26;
	    token_table[k].tok_addr = crt_pos; 	
	    memcpy(&len, crt_pos+1, 4);
	    for (i1 = 0, s = crt_pos+1+4; i1 < len; i1++) {
             	shtlen = strlen(s)+1;
		s += shtlen;
	    }
 	    ttllen  -= 1 + 4+(s - (crt_pos + 1 + 4));
	    crt_pos += 1 + 4+(s - (crt_pos + 1 + 4));
            break;

        case (0x3D): /* AUT_EXEC_ENV */
/*	    if (1) {
		char *s;
		int fd;

		if ((fd = open("record", O_WRONLY)) == 0) {
			fprintf(stderr, "could'nt fopen record\n");
			exit(1);
		}
		if ((write(fd, prec, TOTAL)) != TOTAL){
			fprintf(stderr, "could'nt fwrite record\n");
			exit(1);
		}

		close(fd);
		exit(0);
	    } 
*/
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 6;
	    token_table[k].tok_addr = crt_pos; 	
	    memcpy(&len, crt_pos+1, 4);
	    for (i1 = 0, s = crt_pos+1+4; i1 < len; i1++) {
             	shtlen = strlen(s)+1;
		s += shtlen;
	    }
 	    ttllen  -= 1 + 4+(s - (crt_pos + 1 + 4));
	    crt_pos += 1 + 4+(s - (crt_pos + 1 + 4));
            break;

        case (0x28): 
	    if (text_nb <= 3) {	

		/* accept max. 4 such tokens and skip over the rest  */

	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 28;
	    	token_table[k+text_nb].tok_addr = crt_pos;
	    }  	
	    memcpy(&ushtlen, crt_pos + 1, 2);
 	    ttllen  -= 1 + 2 + ushtlen;
	    crt_pos += 1 + 2 + ushtlen;
	    text_nb++;
            break;

        case (0x13):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 33;
	    token_table[k].tok_addr = crt_pos; 	
 	    ttllen  -= 1 + 2+4;
	    crt_pos += 1 + 2+4;
            break;

        case (0x2A):
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 25;
	    token_table[k].tok_addr = crt_pos; 	
 	    ttllen  -= 1 + 4;
	    crt_pos += 1 + 4;
            break;

        case (0x2B):/* AUT_IP */
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 14;
	    token_table[k].tok_addr = crt_pos; 	
 	    ttllen  -= 1 + 20;
	    crt_pos += 1 + 20;
            break;

        case (0x2E): 
	    	/* k = lookup_tok(tok_id, token_table, TAB_LEN); */
		k = 1;
	    token_table[k].tok_addr = crt_pos; 	
 	    ttllen  -= 1 + 2+2+4; 	/* BUG in BSM doc +2+4; */
	    crt_pos += 1 + 2+2+4;	/* BUG in BSM doc +2+4; */
	   /* while (!*crt_pos) {
		ttllen--;
		crt_pos++;
	    } */
            break;

	default:

		fprintf(stderr, "invalid token id 0x%x\n", tok_id);
		exit(1);		
    }
    } while (ttllen > 0);
#ifdef DEBUG
    fprintf(stdout, "\n");
    fprintf(stdout, "converting record:");
#endif
    for (k = 0; k<TAB_LEN; k++) {
    	if (!token_table[k].tok_addr) 
		continue;
	tok_id = *(token_table[k].tok_addr); 
#ifdef DEBUG
	fprintf(stdout, " 0x%x", tok_id);
#endif
	prec = (void *) token_table[k].tok_addr;
    	switch(tok_id) {
        case (0x2D):
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 1, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec+1+1+4, pbuflen, pbuf);
	    memcpy(&ushtlen, (char *) prec+1+1+4, 2);
            put_bytes(&alpha, N++, ushtlen, (char *) prec+1+1+4+2, pbuflen, pbuf);
            break;
        case (0x31):
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4, pbuflen, pbuf);
            break;
        case (0x21):
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 1, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 1, (char *) prec+1+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 1, (char *) prec+1+1+1, pbuflen, pbuf);
            switch(*(char *) prec+1+1) {
                case (AUR_BYTE):
                    put_bytes(&alpha, N++, *(char *) prec+1+1+1, 
				(char *) prec+1+1+1+1, pbuflen, pbuf);
                    break;
               case (AUR_SHORT):
		    N += 1;
		    memcpy(&ushtlen, (char *) prec +1+1+1, 2);
                    for (i1 = 0; i1 < ushtlen; i1++) {
                        put_bytes(&alpha, N++, 2, (char *) prec+1+1+1+1+i1*2, pbuflen, pbuf);
                    }
                    break;
                case (AUR_INT):
                case (AUR_LONG):
                    N += 5;
		    memcpy(&ushtlen, (char *) prec +1+1+1, 2);
                    for (i1 = 0; i1 < ushtlen; i1++) {
                        put_bytes(&alpha, N++, 4, (char *) prec+1+1+1+1+i1*4, pbuflen, pbuf);
                    }
                   break;
            }
            break;

        case (0x52):
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4, pbuflen, pbuf);
            break;

        case (0x11):
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec+1+4+4, pbuflen, pbuf);
	    memcpy(&ushtlen, (char *) prec+1+4+4, 2);
            put_bytes(&alpha, N++, ushtlen, (char *) prec+1+4+4+2, pbuflen, pbuf);
            break;
        case (0x34):
	    N = token_table[k].base_id;
            for (i1 = 0; i1 < 16; i1++) {
                put_bytes(&alpha, N++, 4, (char *) prec+1+i1*4, pbuflen, pbuf);
            }
            break;
        case (0x14):
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 1, (char *) prec+1+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec+1+4+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec+1+4+1+2, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+1+2+2, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+1+2+2+4, pbuflen, pbuf);
            break;
        case (0x32):
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4, pbuflen, pbuf);
            for (i1 = 0; i1 < 4; i1++) {
                put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4+4 + i1*4, pbuflen, pbuf);
            }
            break;
        case (0x22): /* AUT_IPC */
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 1, (char *) prec + 1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec + 1+1, pbuflen, pbuf);
            break;
        case (0x2C): /* == AUT_IPORT (Internet port address) */
	    N = token_table[k].base_id;
    	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec + 1, pbuflen, pbuf);
            break;
        case (0x00):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec + 1, pbuflen, pbuf);
	    memcpy(&ushtlen, (char *) prec + 1, 2);
            put_bytes(&alpha, N++, ushtlen+1, (char *) prec + 1 +2, pbuflen, pbuf);
            break;
        case (0x29):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec + 1, pbuflen, pbuf);
	    memcpy(&ushtlen, (char *) prec + 1, 2);
            put_bytes(&alpha, N++, ushtlen, (char *) prec + 1 +2, pbuflen, pbuf);
            break;
        case (0x23):
	    N = token_table[k].base_id;
            put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec + 1, pbuflen, pbuf);
	    memcpy(&ushtlen, (char *) prec + 1, 2);
	    if (!strncmp((char *) prec + 1 +2, "///", 3)) 
            put_bytes(&alpha, N++, ushtlen-2, (char *) prec + 1 +2+2, pbuflen, pbuf);
	    else 
            put_bytes(&alpha, N++, ushtlen, (char *) prec + 1 +2, pbuflen, pbuf);
            break;
        case (0x26):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4+4+4, pbuflen, pbuf);
            break;
        case (0x27):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 1, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+1, pbuflen, pbuf);
            break;
        case (0x25):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4, pbuflen, pbuf);
            break;
        case (0x2F):
	    N = token_table[k].base_id;
            put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            break;
        case (0x24):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+4+4+4+4+4+4+4+4, pbuflen, pbuf);
            break;
         case (0x3B): /* AUT_NEWGROUPS */
	    N = token_table[k].base_id;

     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec+1, pbuflen, pbuf);
	    memcpy(&ushtlen, (char *) prec+1, 2);
	    for (i1 = 0; i1 < ushtlen; i1++) {
             	put_bytes(&alpha, N++, 4, (char *) prec+1+2+i1*4, pbuflen, pbuf);
	    }
            break;
       case (0x3C): /* AUT_EXEC_ARGS */
	    N = token_table[k].base_id;

     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
	    memcpy(&len, (char *) prec+1, 4);
	    for (i1 = 0, s = (char *) prec+1+4; i1 < len; i1++) {
             	shtlen = put_string(&alpha, N++, s, pbuflen, pbuf);
		s += shtlen;
	    }
            break;

        case (0x3D): /* AUT_EXEC_ENV */
	    N = token_table[k].base_id;

     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1, pbuflen, pbuf);
	    memcpy(&len, (char *) prec+1, 4);
            shtlen=put_stringMulti(&alpha, N++, len, (char *) prec+1+4, pbuflen, pbuf);
            break;

        case (0x28):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec + 1, pbuflen, pbuf);
	    memcpy(&ushtlen, (char *) prec + 1, 2);
            put_bytes(&alpha, N++, ushtlen, (char *) prec + 1 +2, pbuflen, pbuf);
            break;
        case (0x13):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 2, (char *) prec + 1, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec + 1+2, pbuflen, pbuf);
            break;
        case (0x2A):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec + 1, pbuflen, pbuf);
            break;
        case (0x2B):
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
            put_bytes(&alpha, N++, 20, (char *) prec + 1, pbuflen, pbuf);
            break;
        case (0x2E): /* AUT_SOCKET */
	    N = token_table[k].base_id;
     	    put_bytes(&alpha, N++, 1, (char *) prec, pbuflen, pbuf);
     	    put_bytes(&alpha, N++, 2, (char *) prec+1, pbuflen, pbuf);
     	    put_bytes(&alpha, N++, 2, (char *) prec+1+2, pbuflen, pbuf);
     	    put_bytes(&alpha, N++, 4, (char *) prec+1+2+2, pbuflen, pbuf);
         /*   
		BUG in BSM doc :

	    put_bytes(&alpha, N++, 2, (char *) prec+1+2+2+4, pbuflen, pbuf);
            put_bytes(&alpha, N++, 4, (char *) prec+1+2+2+4+2, pbuflen, pbuf);
	 */
            break;
	default:

		fprintf(stderr, "invalid token id 0x%x\n", tok_id);
		exit(1);
		

    }
    } 
#ifdef DEBUG
    fprintf(stdout, "\n");
#endif


    *((int *) *pbuf) = alpha - *pbuf;
}

 
