/*
 :******************************************************************
 :* MODULE-DESCRIPTION
 :*
 :* TITLE         (/ Lexical Analyser /)
 :*
 :* NAME          scanLib.c
 :*
 :* COPYRIGHT     (C)   SIEMENS NIXDORF INFORMATIONSSYSTEME AG  1992
 :*                     ALL RIGHTS RESERVED
 :*
 :* MEMORY-CLASS  6
 :*
 :* PROG-CONVENTS (/ <internal conventions (instance, names,
 :*                  external data, ...). None -> empty text>     /)
 :*
 :* PROCEDURE     getsymb
 :* REMARKS       (/ reading of a token in the input stream              /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     isRcomment
 :* REMARKS       (/ check if this is the end of a comment              /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     getComment
 :* REMARKS       (/ skip a comment and return the next token           /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     issymbol
 :* REMARKS       (/ check if this is a specified symbol                /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     get_spec
 :* REMARKS       (/ reading of a special symbol in the input stream     /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     get_identifier
 :* REMARKS       (/ reading of an identifier                           /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     takeofchar
 :* REMARKS       (/ skip alphanumeric characters                       /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     internRepr
 :* REMARKS       (/ build the internal repres. of an identifier        /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     skipspaces
 :* REMARKS       (/ skip spaces in the input stream            /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     get_char
 :* REMARKS       (/ reading of a legal character in the input steam /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     get_char1
 :* REMARKS       (/ reading of a character in the input stream /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     isBadChar
 :* REMARKS       (/ check if this is a legal character /)
 :* ENTRY         EXTERNAL
 :*
 :*-----------------------------------------------------------------
 :*
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz    , FUNDP /)
 :*   UPDATE      (/ original issue /)
 :*
 :* END-MODULE-DESCRIPTION scanLib.c
 :******************************************************************
*/
#include <stdio.h>
#include <ctype.h>      /* declares function for testing characters  */
#include "specTab.h"
#define specsymbnbre  5 /* nbre of special symbols */
static char *specsymbtab[] = {"integer", "ref", "string", "undef", "val",
		       "(", ")", ",", ":", ";", "."};

#define tabchar '\t'
#define eoline  '\n'
#define space  ' '
#define quote  '\''
int nextchar;           /* last read character  */
short int more_char;      /* = EOF */
static char resymb[MAXS];
/*
 :******************************************************************
 :* UNIT-BODY scanLib.getsymb
 :*
 :* TITLE         (/ reading of a token in the input stream      /)
 :*
 :* PURPOSE       reading of the next token in the input stream and
 :*               storing its internal representation.
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA
 :*       Global Data:  (/
 :*                       - specsymbtab: Table of special symbol of the
 :*                                      rule based language.
 :*                       - symb       : holds the internal representation
 :*                                      of the read symbol.
 :*                       - nextchar   : the last character being read.
 :*                       - more_char  : is TRUE when end of file is.
 :*                       - LineNbre   : the line number.
 :*                                                                  /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre (/ - fp      a valid file pointer
 :*          - Let s be the current input stream.
 :*                s = s1 t
 :*            where
 :*                  - s1 is a possibly empty sequence of spaces (blanks, eol
 :*                    and tabs);
 :*                  - there exist a greatest prefixe t1 of t such that:
 :*                      t = t1 u
 :*                    where t1 is a symbol and u is any character sequenece
 :*       /)
 :*                                                             END-Pre
 :*   Post   (/ - the input stream is u;
 :*             - nextchar contains the first character in u;
 :*             - symb contains the internal representation of t1.  /)
 :*             - LineNbre contains the current line number
 :*             - more_char is TRUE if end of file is reached.
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.getsymb
 :******************************************************************
*/

getsymb(fp)
FILE *fp;
{
	int index;

	skipspaces(fp);
	if (more_char == 0)
	   symb.ibsymb = nosymbx;
	else if (isalpha(nextchar))
		get_identifier(fp);
	else if ((index = issymbol(nextchar)) >= 0) {
                symb.svalue[0] = nextchar;
                symb.svalue[1] = '\0';
		symb.ibsymb = (enum index) index;
		get_char(fp);
	} else if (nextchar == '/')
		getComment(fp);
	else {
		symb.ibsymb = errsymbx;
		get_char(fp);
	}
/*        printsymb(); */

}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.isRcomment
 :*
 :* TITLE         (/ checking the end of comment /)
 :*
 :* PURPOSE       check if the input stream starts with the sequence
 :*               of two characters "star slash" i.e. the end of a comment.
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA
 :*       Global Data:  (/
 :*                        nextchar   : the last character being read.
 :*                                                                  /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - fp      a valid file pointer          /)
 :*                                                             END-Pre
 :*   Post   (/ - if the input stream is of the form:
 :*                    <star slash><rest>
 :*                 where
 :*                 - star is '*' and slash is '/'
 :*                 - <rest> is the rest of the input stream.
 :*                the effect is
 :*                 - nextchar contains the first character in <rest>;
 :*                 - LineNbre contains the current line number;
 :*                 - the input stream is <rest>
 :*                 - the returned value = 1
 :*                otherwise the returned value = 0 and the input stream is
 :*                 unchanged.
 :*                                                                   /)
 :*                                                             END-Post
 :*   RETURNS  "int" (/ = 1 if "star slash" is a prefixe of the input stream
 :*                     = 0 otherwise                              /)
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.isRcomment
 :******************************************************************
*/
isRcomment(fp)  /* returns 1 if star slash is a prefixe of
		   the remaining char. seq. return 0 otherwise */

FILE *fp;
{
    if (nextchar == '*') {
	get_char(fp);
	if (nextchar == '/') {
	    get_char(fp);
	    return(1);
	} else {
	    return (0);
	}
    } else {
	return(0);
    }
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.getComment
 :*
 :* TITLE         (/ reading of a comment in the input stream      /)
 :*
 :* PURPOSE       skip the entire current comment and then read the
 :*               the symbol after this comment.
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA
 :*       Global Data:  (/
 :*                       - symb       : holds the internal representation
 :*                                      of the read symbol.
 :*                       - nextchar   : the last character being read.
 :*                       - more_char  : is TRUE when end of file is.
 :*                       - LineNbre   : the line number.
 :*                                                                  /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre (/ - fp      a valid file pointer
 :*          - nextchar = '/'
 :*       /)
 :*                                                             END-Pre
 :*   Post   (/ - if the input stream is of the form:
 :*                    <star><comment_body><star slash><whites><token><rest>
 :*                 where
 :*                 - star is '*' and slash is '/'
 :*                 - comment_body is the character sequence of the comment;
 :*                 - whites is any sequence of spaces (blanks, eol and tabs)
 :*                 - token is a symbol;
 :*                 - rest is the rest of the input stream.
 :*                the effect is
 :*                 - nextchar contains the first character in <rest>;
 :*                 - symb contains the internal representation of <token>;
 :*                 - LineNbre contains the current line number;
 :*                 - more_char is TRUE if end of file is reached;
 :*                 - the input stream is <rest>
 :*                otherwise 'symb' contains the internal repres.
 :*                of the error symbol.
 :*                                                                   /)
 :*                                                             END-Post
 :*   RETURNS  "int" (/ if this is a legal comment = 1 else = 0 /)
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.getComment
 :******************************************************************
*/
getComment(fp)  /* skip a comment and get the next token */
FILE *fp;
{
    get_char(fp);
    if (nextchar == '*') {
	get_char1(fp);
	while (more_char && !isRcomment(fp)) {
	    get_char1(fp);
	}
	if (more_char) {
	    getsymb(fp);
	} else {
	    symb.ibsymb = errsymbx;
	}
    } else {
	symb.ibsymb = errsymbx;
    }
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.issymbol
 :*
 :* TITLE         (/ checking if the current character is a member
 :*                  of a given list.   /)
 :*
 :* PURPOSE       checking if the current character is a member of a
 :*               subset of the special symbol table. If this is the
 :*               case return its index; otherwise return -1.
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*                                                               /)
 :*   DATA
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - c    a given character  /)
 :*                                                             END-Pre
 :*   Post   (/ - if c in {'+', '-', '*', '(', ')', '=', ':', ';', ',', '.'}
 :*                 return index
 :*                   where specsymbtab[index] = c;
 :*               otherwise returned -1.
 :*                                                                   /)
 :*                                                             END-Post
 :*   RETURNS  "int" (/ = index of the character in 'specsymbtab' if a member
 :*                     = -1 otherwise                              /)
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.issymbol
 :******************************************************************
*/
issymbol(c)
char c;
{
	switch (c) {
		case '(': return(lparenthesis);
		case ')': return(rparenthesis);
		case ':': return(colon);
		case ';': return(semicolon);
		case ',': return(comma);
		case '.': return(period);
		default : return (-1);
	}
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.get_identifier
 :*
 :* TITLE         (/ reading of an identifier    /)
 :*
 :* PURPOSE       (/ read of the identifier that begins the input
 :*                  stream and store its internal repres.
 :*                                                           /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*
 :*      Invariant:
 :*
 :*               - c0 c1 ... ci <stream>
 :*               - 0 <= i < ident_lg
 :*               - <stream> is the current input stream
 :*               - for j=0 .. i
 :*                               cj is a letter, a digit or '_'
 :*                               resymb[j] = cj
 :*                                                                 /)
 :*
 :*   DATA       (/
 :*                       - specsymbtab: Table of special symbol of the
 :*                                      rule based language.
 :*                       - symb       : holds the internal representation
 :*                                      of the read symbol.
 :*                       - resymb     : last read symbol.
 :*                       - nextchar   : the last character being read.
 :*                       - more_char  : is TRUE when end of file is.
 :*                       - LineNbre   : the line number.
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - fp    a valid file pointer.
 :*             - nextchar is alphanumeric
 :*             - the input stream is of the form:
 :*                      <identifier><rest>
 :*                                                                      /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - symb contains the internal repres. of <identifier1>
 :*               where
 :*                     <identifier1> is <identifier> truncated to its
 :*                     'ident_lg' first characters;
 :*             - the input stream is <rest>                          /)
 :*
 :*                                                             END-Post
 :*
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.get_identifier
 :******************************************************************
*/
get_identifier(fp)
FILE *fp;
{
	int i;

	i = 0;
	while ((isalnum(nextchar) || nextchar == '_') &&
		  i < ident_lg - 1   ) {   /* identif_lg --> ident_lg */
		resymb[i] = nextchar;
		i++;
		get_char(fp);
	}
	resymb[i] = '\0';
	takeofchar(fp);
	internRepr();
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.takeofchar
 :*
 :* TITLE         (/ skip alphanumeric characters   /)
 :*
 :* PURPOSE       (/ read all successive alphanumeric characters from
 :*                  the input stream i.e. search for identifier end.
 :*                                                           /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/                                              /)
 :*
 :*      Invariant:
 :*
 :*               - c0 c1 ... ci <stream>
 :*               - <stream> is the current input stream
 :*               - for j=0 .. i
 :*                               cj is a letter, a digit or '_'
 :*                                                                 /)
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - fp    a valid file pointer.
 :*             - nextchar is alphanumeric or an '_'
 :*             - the input stream is of the form:
 :*                      <alphanum><rest>
 :*               where <alphanum> is a possibly empty sequence of                                                       /)
 :*                                letters digits and '_'
 :*                     <rest>     is a sequence not beginning with a letter
 :*                                a digit or '_'
 :*                                                            /)
 :*
 :*                                                             END-Pre
 :*
 :*   Post   (/ - the input stream is <rest>                          /)
 :*                                                             END-Post
 :*
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.takeofchar
 :******************************************************************
*/
takeofchar(fp)
FILE *fp;
{
	while (isalpha(nextchar) || isdigit(nextchar) || nextchar == '_' )
		get_char(fp);
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.internRepr
 :*
 :* TITLE         (/ build the internal repres. of an identifier      /)
 :*
 :* PURPOSE       (/ given the external represenation of an identifier,
 :*                  generate the internal repres.
 :*                                                           /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*   DATA       (/
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - fp    a valid file pointer.
 :*             - resymb is a string representing the identifier.
 :*                                                            /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - - symb contains the internal representation of resymb. /)
 :*                                                             END-Post
 :*
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.internRepr
 :******************************************************************
*/
internRepr()
{
	int index;

	if ((index = isspecsymb(resymb)) >= 0) {
		symb.ibsymb = (enum index) index;
	} else /* identifier */ {
		symb.ibsymb = identifierx;
	}
	strcpy(symb.svalue, resymb);
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.isspecsymb
 :*
 :* TITLE         (/ check if this is a special symbol           /)
 :*
 :* PURPOSE       (/ search for the given symbol in the table of
 :*                  special symbols specsymbtab, return its index if
 :*                  found -1 otherwise
 :*                                                           /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*   DATA       (/
 :*                       - specsymbtab: Table of special symbol of the
 :*                                      rule based language.
 :*                       - symb       : holds the internal representation
 :*                                      of the read symbol.
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - s is a string
 :*                                                            /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - if there exist i such that the string specsymbtab[i]
 :*               is equal to s return i otherwise return -1
 :*
 :*                                                             END-Post
 :*
 :*   RETURNS  "int" (/ if s is in specsymbtab the returned value is
 :*                        the index of s in specsymbtab
 :*                     otherwise the returned value is -1
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.isspecsymb
 :******************************************************************
*/
isspecsymb(s)
char s[MAXS];
{
	int low, high, mid, k;

	low = 0;
	high = specsymbnbre - 1 ;
	while (low <= high) {
		mid = (low + high) / 2;
		if ((k = strcmp(s, specsymbtab[mid])) < 0)
			high = mid - 1;
		else if (k > 0)
			low = mid + 1;
		else /* s is a special symbol */
			return (mid);
	}
	return (-1);
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.skipspaces
 :*
 :* TITLE         (/ skip spaces in the input stream       /)
 :*
 :* PURPOSE       (/ skipping all white characters (blank, tab and eol)
 :*                  in the input stream
 :*                                                           /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*   DATA       (/
 :*                      - nextchar : the last read character
 :*                      - morechar : = EOF
 :*                      - LineNbre : the current line number
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - fp    a valid file pointer.
 :*             - the input stream is in the form:
 :*                    <spaces><rest>
 :*               where <spaces> is a possibly empty sequence of blanks,
 :*                              tabs and end of lines
 :*                     <rest> is a sequence not beginning with a blank,
 :*                            a tab or an end of line
 :*                                                            /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - the input stream is <rest>
 :*                                                             END-Post
 :*
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.skipspaces
 :******************************************************************
*/
skipspaces(fp)
FILE *fp;
{
    while (more_char == 1 &&
	  (nextchar == space || nextchar == eoline || nextchar == tabchar)) {
		get_char(fp);
    }
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.get_char
 :*
 :* TITLE         (/ reading of a legal character in the input stream /)
 :*
 :* PURPOSE       (/ reading of the next legal character in the input
 :*                  stream ignoring bad ones.
 :*                                                           /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*   DATA       (/
 :*                      - nextchar : the last read character
 :*                      - LineNbre : the current line number
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - fp    a valid file pointer.
 :*             - the input stream is of the form:
 :*                   <badCharSeq> c <rest>
 :*               where <badCharSeq> is a possibly empty sequence of
 :*               illegal characters.
 :*                                                            /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - the input stream is <rest>
 :*             - nextchar = c
 :*             - n error messages are output to stdout where n is the
 :*               length of <badCharSeq>
 :*             - morechar and LineNbre are updated accordingly    /)
 :*                                                             END-Post
 :*
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.get_char
 :******************************************************************
*/
get_char(fp)
FILE *fp;
{
    get_char1(fp);
    while (isBadChar()) {
	printf("lexical      %d: bad character: %x\n",
		LineNbre, nextchar);
	get_char1(fp);     /* ignore bad characters */
    }
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.get_char1
 :*
 :* TITLE         (/ reading of a character in the input stream /)
 :*
 :* PURPOSE       (/ read the next character in the input stream.
 :*                  increment the line number if end of line.
 :*                                                           /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/             /)
 :*   DATA       (/
 :*                      - nextchar : the last read character
 :*                      - morechar : = EOF
 :*                      - LineNbre : the current line number
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - fp    a valid file pointer.
 :*             - the input stream is of the form:
 :*                      c <rest>
 :*                                                            /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - if c is end of file
 :*                   morechar = 0
 :*               otherwise morechar = 1
 :*               if c is end of line LineNbre is incremented by 1.
 :*             - the input stream is <rest>
 :*
 :*                                                             END-Post
 :*
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.get_char1
 :******************************************************************
*/
get_char1(fp)
FILE *fp;
{
	if ((nextchar = getc(fp)) == EOF) {
	    more_char = 0;
	}
	else {
	    more_char = 1;
	    if (nextchar == eoline) LineNbre++;
	}
}
symbolClass(index)
int index;
{
	int aux1, aux2;

	aux1 = intx;
	aux2 = period;
	if (index >= aux1 && index <= aux2) 
             return(specsymb);
	else switch (index) {
		case nosymbx    : return(nosymb);
		case errsymbx   : return(errsymb);
		default         : return(0);
	     }
}
/*
 :******************************************************************
 :* UNIT-BODY scanLib.isBadChar
 :*
 :* TITLE         (/ check if this is a legal character    /)
 :*
 :* PURPOSE       (/ checking if the last read character is a legal one /)
 :*
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-11-03
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/  /)
 :*   DATA       (/
 :*                  - nextchar   : the last character being read.
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/ Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/
 :*                                                            /)
 :*                                                             END-Pre
 :*
 :*   Post   (/ - if nextchar in
 :*                    A = {'+', '-', '*', '(', ')', '>', '<', '!', '=',
 :*                         ':', ';', ',', '/', '.', '\', '_', ''', '%' ' '}
 :*                 or nextchar is a letter, a digit, a tab, an end of line
 :*                 or an end of file
 :*               return 0
 :*               return 1 otherwise
 :*
 :*                                                             END-Post
 :*   RETURNS  "int" (/ =0 if nextchar is among the above list
 :*                     =1 otherwise                              /)
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY scanLib.isBadChar
 :******************************************************************
*/
isBadChar()
{
    if ( isalpha(nextchar) || isdigit(nextchar) ||
	 nextchar == '+' || nextchar == '-' || nextchar == '*' ||
	 nextchar == '(' || nextchar == ')' || nextchar == '>' ||
	 nextchar == '<' || nextchar == '!' || nextchar == '=' ||
	 nextchar == ':' || nextchar == ';' || nextchar == ',' ||
	 nextchar == '/' || nextchar == '.' || nextchar == '\\' ||
	 nextchar == '_' ||
	 nextchar == quote || nextchar == tabchar || nextchar == eoline ||
	 nextchar == space || nextchar == EOF || nextchar == '%')

	 return(0);
    else
	 return(1);
}


